﻿using System.Collections.ObjectModel;
using CommunityToolkit.Mvvm.ComponentModel;
using CommunityToolkit.Mvvm.Input;
using Microsoft.Extensions.Logging;
using StudiApp.Interfaces;
using StudiApp.Models;
using StudiApp.Services.Converters;

namespace StudiApp.ViewModels;

[QueryProperty(nameof(Mark), "Mark")]
public partial class MarksEditViewModel : ObservableObject, IQueryAttributable
{
    private readonly ILogger<MarksEditViewModel> _logger;
    private readonly IRepository<Mark, int> _markRepository;
    private readonly IRepository<Course, int> _courseRepository;

    [ObservableProperty] private Mark _mark;
    [ObservableProperty] private string _percentage;
    [ObservableProperty] private ObservableCollection<Course> _courses;

    [ObservableProperty] private bool _courseHasError;
    [ObservableProperty] private bool _descriptionHasError;
    [ObservableProperty] private bool _markHasError;
    [ObservableProperty] private bool _percentageHasError;

    private string _errorMessage;

    public MarksEditViewModel(
        ILogger<MarksEditViewModel> logger,
        IRepository<Mark, int> markRepository,
        IRepository<Course, int> courseRepository)
    {
        _logger = logger;
        _markRepository = markRepository;
        _courseRepository = courseRepository;
        _courses = new ObservableCollection<Course>();

        LoadCourses();

        _logger.LogInformation($"Initialized {nameof(MarksEditViewModel)}");
    }

    public void ApplyQueryAttributes(IDictionary<string, object> query)
    {
        Mark = (query[nameof(Mark)] as Mark)!;
        Percentage = PercentageConverter.ConvertToString(Mark.Percentage);
        _logger.LogInformation($"Edit Page for Mark {Mark.Id} loaded");
    }

    [RelayCommandAttribute]
    public async void Save()
    {
        Mark.Percentage = ConvertPercentageToDouble(Percentage);

        var isInputValid = ValidateInput();

        if (isInputValid)
        {
            _logger.LogInformation("Mark name: " + Mark.Description);
            
            await _markRepository.Update(Mark);

            _logger.LogInformation("Altered Mark Saved");

            await Shell.Current.GoToAsync("..");
        }
        else
        {
            // TODO: Implement dialog service
        }
    }

    private async void LoadCourses()
    {
    
        List<Course> courses = await _courseRepository.GetAll();

        foreach (var course in courses)
        {
            Courses.Add(course);
        }
        
    }

    private double ConvertPercentageToDouble(string percentage)
    {
        return PercentageConverter.ConvertToDouble(percentage);
    }

    private bool ValidateInput()
    {
        _errorMessage = string.Empty;

        return ValidateCourse() &&
               ValidateDescription() &&
               ValidateMarkValue() &&
               ValidatePercentage();
    }

    private bool ValidateCourse()
    {
        if (Mark.Course != null)
        {
            CourseHasError = false;
            return true;
        }

        CourseHasError = true;
        _errorMessage += "A course must be chosen\n";
        return false;
    }

    private bool ValidateDescription()
    {
        if (Mark.Description is { Length: <= 250 and > 1 })
        {
            DescriptionHasError = false;
            return true;
        }

        DescriptionHasError = true;
        _errorMessage += "Description between 1 - 250 chars\n";
        return false;

    }

    private bool ValidateMarkValue()
    {
        if (Mark.Value is >= 1.0 and <= 6.0)
        {
            MarkHasError = false;
            return true;
        }

        MarkHasError = true;
        _errorMessage += "Mark value out of range, must be between 1.0 and 6.0\n";
        return false;
    }

    private bool ValidatePercentage()
    {
        if (Mark.Percentage is > 0 and <= 1 && ValidatePercentageOverAllMarks())
        {
            PercentageHasError = false;
            return true;
        }

        if (!ValidatePercentageOverAllMarks())
        {
            MarkHasError = true;
            _errorMessage += "Percentage is above 100% over all marks of this course";
            return false;
        }

        MarkHasError = true;
        _errorMessage += "Percentage out of range, must be in range 0.1 up to 1.0";
        return false;
    }

    private bool ValidatePercentageOverAllMarks()
    {
        var overallCourseMarkPercentage = Mark.Course.Marks.Sum(mark => mark.Percentage);

        return (overallCourseMarkPercentage) <= 1.0;
    }

}

