﻿using System.Collections.ObjectModel;
using CommunityToolkit.Mvvm.ComponentModel;
using CommunityToolkit.Mvvm.Input;
using Microsoft.Extensions.Logging;
using StudiApp.Interfaces;
using StudiApp.Models;
using StudiApp.Services.Converters;
using StudiApp.Views;

namespace StudiApp.ViewModels;
public partial class MarksOverviewViewModel : ObservableObject
{
    private readonly ILogger<MarksOverviewViewModel> _logger;

    private readonly IRepository<Semester, int> _semesterRepository;
    private readonly IRepository<Course, int> _courseRepository;
    private readonly IRepository<Mark, int> _markRepository;

    [ObservableProperty]
    private double _totalAverageMark;

    [ObservableProperty]
    private int _totalEcts;

    [ObservableProperty]
    private double _semesterAverageMark;

    [ObservableProperty]
    private int _semesterEcts;

    [ObservableProperty] private ObservableCollection<Mark> _marks;

    [ObservableProperty] 
    private ObservableCollection<Semester> _semesters;

    [ObservableProperty] 
    private Semester _selectedSemester;

    public MarksOverviewViewModel(ILogger<MarksOverviewViewModel> logger,
        IRepository<Semester, int> semesterRepository,
        IRepository<Course, int> courseRepository,
        IRepository<Mark, int> markRepository)
    {
        _semesterRepository = semesterRepository;
        _courseRepository = courseRepository;
        _markRepository = markRepository;
        _logger = logger;

        InitializeSemesterData();
        LoadTotalOverview();
    }

    [RelayCommand]
    public void OnAppearing()
    {
        LoadSemesterOverview();
    }

    private async void InitializeSemesterData()
    {
        //CreateMockData(); // TODO: remove this when not needed anymore
        var result = await _semesterRepository.GetAll();
        Semesters = new ObservableCollection<Semester>(result);
        SelectedSemester = Semesters[0]; // TODO: get actual current 
    }

    private async void LoadSemesterOverview()
    {
        // TODO: if SelectedSemester is null, show text 'Select a semester'
        var result = await _courseRepository.GetAll();
        var semesterCourses = result.Where(c => c.SemesterId == SelectedSemester.Id).ToList();
        var semesterMarks = GetMarksFromCourses(semesterCourses);

        Marks = new ObservableCollection<Mark>(semesterMarks);
        SemesterEcts = GetObtainedEcts(semesterCourses);
        SemesterAverageMark = AverageMarkConverter.MarksToAverage(semesterMarks);
    }

    private void LoadTotalOverview()
    {
        _courseRepository.GetAll().ContinueWith(t => TotalEcts = GetObtainedEcts(t.Result));
        _markRepository.GetAll().ContinueWith(t => TotalAverageMark = AverageMarkConverter.MarksToAverage(t.Result));
    }

    private List<Mark> GetMarksFromCourses(List<Course> courses)
    {
        return courses.SelectMany(c => c.Marks).ToList();
    }

    partial void OnSelectedSemesterChanged(Semester value)
    {
        LoadSemesterOverview();
    }

    private int GetObtainedEcts(IEnumerable<Course> courses)
    {
        var passedCourses = courses.Where(course => course.IsPassed);
        return passedCourses.Sum(course => course.Ects);
    }

    [RelayCommand]
    public async Task EditMark(Mark mark)
    {
        await Shell.Current.GoToAsync(nameof(MarksEditView),
            new Dictionary<string, object> { {"Mark", mark} });
        _logger.LogInformation($"Navigating to {nameof(MarksEditView)} (edit mark)");
    }

    [RelayCommand]
    public async Task NewMark()
    {
        await Shell.Current.GoToAsync(nameof(MarksCreateView));
        _logger.LogInformation($"Navigating to {nameof(MarksCreateView)} (create new mark)");
    }


    // TODO: remove this function when not needed anymore
    private async void CreateMockData()
    {
        var semester1 = new Semester()
        {
            DateRange = new DateRange() { Start = new DateTime(2032, 1, 1), End = new DateTime(2023, 6, 1) },
            Name = "Semester 1"
        };
        var semester2 = new Semester()
        {
            DateRange = new DateRange() { Start = new DateTime(2032, 6, 2), End = new DateTime(2023, 12, 1) },
            Name = "Semester 2"
        };

        semester1 = await _semesterRepository.Create(semester1);
        semester2 = await _semesterRepository.Create(semester2);

        var course1 = new Course() { Name = "Course 1", Ects = 4, IsPassed = true, Semester = semester1, SemesterId = semester1.Id };
        var course2 = new Course() { Name = "Course 2", Ects = 2, IsPassed = true, Semester = semester1, SemesterId = semester1.Id };
        var course3 = new Course() { Name = "Course 3", Ects = 4, IsPassed = true, Semester = semester2, SemesterId = semester2.Id };

        course1 = await _courseRepository.Create(course1);
        course2 = await _courseRepository.Create(course2);
        course3 = await _courseRepository.Create(course3);

        await _markRepository.Create(new Mark { Value = 4, Description = "Exam 1", Percentage = 0.5, Course = course1, CourseId = course1.Id });
        await _markRepository.Create(new Mark { Value = 4, Description = "Exam 2", Percentage = 0.5, Course = course1, CourseId = course1.Id });
        await _markRepository.Create(new Mark { Value = 4.5, Description = "Exam 1", Percentage = 0.5, Course = course2, CourseId = course2.Id });
        await _markRepository.Create(new Mark { Value = 5.5, Description = "Exam 2", Percentage = 0.5, Course = course2, CourseId = course2.Id });
        await _markRepository.Create(new Mark { Value = 1, Description = "Testat 1", Percentage = 0.5, Course = course3, CourseId = course3.Id });
        await _markRepository.Create(new Mark { Value = 4.5, Description = "Testat 2", Percentage = 0.3, Course = course3, CourseId = course3.Id });
        await _markRepository.Create(new Mark { Value = 4, Description = "Final Exam", Percentage = 0.2, Course = course3, CourseId = course3.Id });
    }
}
