﻿using CommunityToolkit.Mvvm.ComponentModel;
using System.ComponentModel;
using StudiApp.Models;
using Syncfusion.Maui.Scheduler;
using System.Collections.ObjectModel;
using CommunityToolkit.Mvvm.Input;
using StudiApp.Interfaces;

namespace StudiApp.ViewModels;

public partial class TimetableViewModel : ObservableObject
{
    private const int WeekdayLength = 2;
    private const string SemesterFormatting = "yyyyMMddTHHmmssZ";
    private const int TodoBeforeDue = -20;

    [ObservableProperty]
    private ObservableCollection<SchedulerAppointment> _appointments;


    private readonly IRepository<Lecture, int> _lectureRepository;
    private readonly IRepository<ToDo, int> _todoRepository;
    private readonly IRepository<Semester, int> _semesterRepository;

    public TimetableViewModel(IRepository<Lecture, int> lectureRepository,
        IRepository<ToDo, int> todoRepository,
        IRepository<Semester, int> semesterRepository)
    {
        _lectureRepository = lectureRepository;
        _todoRepository = todoRepository;
        _semesterRepository = semesterRepository;
    }

    private async Task<Semester> GetCurrentSemester()
    {
        var semesters = await _semesterRepository.GetAll();
        return Semester.GetCurrent(semesters);
    }

    private async Task<ObservableCollection<SchedulerAppointment>> GetAppointments()
    {
        var lectures = await _lectureRepository.GetAll();
        var appointments = new ObservableCollection<SchedulerAppointment>();
        foreach (var lecture in lectures)
        {
            appointments.Add(AppointmentFromLecture(lecture));
        }

        var todos = await _todoRepository.GetAll();
        foreach (var todo in todos)
        {
            appointments.Add(AppointmentFromTodo(todo));
        }

        return appointments;
    }

    private SchedulerAppointment AppointmentFromLecture(Lecture lecture)
    {
        // test numbers, will come from semester later
        var semesterStart = new DateTime(2023, 1, 1);
        var semesterEnd = new DateTime(2023, 6, 1);
        
        var semesterEndStr = semesterEnd.ToString(SemesterFormatting);
        var weekday = lecture.DayTime.DayOfWeek;
        var weekdayStr = weekday.ToString().Substring(0, WeekdayLength).ToUpper();

        var reccurenceRule = $"FREQ=WEEKLY;INTERVAL=1;UNTIL={semesterEndStr};BYDAY={weekdayStr}";

        var firstLecture = GetNextWeekday(semesterStart, weekday);
        var firstLectureStart = firstLecture.AddHours(lecture.DayTime.StartTime.Hour).AddMinutes(lecture.DayTime.StartTime.Minute);
        var firstLectureEnd = firstLecture.AddHours(lecture.DayTime.EndTime.Hour).AddMinutes(lecture.DayTime.EndTime.Minute);
        return new SchedulerAppointment()
        {
            Id = lecture.Id,
            StartTime = firstLectureStart,
            EndTime = firstLectureEnd,
            Subject = lecture.Name,
            RecurrenceRule = reccurenceRule,
        };
    }

    private SchedulerAppointment AppointmentFromTodo(ToDo todo)
    {
        return new SchedulerAppointment()
        {
            Id = todo.Id,
            StartTime = todo.Due.AddMinutes(TodoBeforeDue),
            EndTime = todo.Due,
            Subject = todo.Name
        };
    }

    private DateTime GetNextWeekday(DateTime date, DayOfWeek weekday)
    {
        var days = ((int) weekday - (int) date.DayOfWeek + 7) % 7;
        return date.AddDays(days);
    }

    [RelayCommand]
    public void OnAppearing()
    {
        GetAppointments().ContinueWith(at => Appointments = at.Result);
    }
}
